"""
"""
from logging import getLogger
from naludaq.controllers.trigger.hdsocv2 import TriggerControllerHdsocv2

logger = getLogger("naludaq.trigger_controller_CODLIR_Eval")


class TriggerControllerCodlir(TriggerControllerHdsocv2):
    def __init__(self, board):
        super().__init__(board)

    def _set_trigger_thresholds(self, values: dict[int, int]):
        """Set the trigger values for the individual channels using the values
        stored in the board object.
        """
        return
        values = values or self.values
        self._validate_thresholds_or_raise(values)
        self._set_tsel_enabled()
        for channel, threshold_value in values.items():
            register_name = self._trig_thresh_regname.format(channel)
            if (
                self.board.registers["analog_registers"][register_name]["value"][0]
                != threshold_value
            ):
                self._write_analog_register(register_name, threshold_value)
